/********************************************************************************

 Programmed by SEUNG-JO BAE 

  - Goal: PACK

  - Algorithm: Seungjo Bae & Sanjay Ranka

  - Input: 

  - Output:

  - Version 1.0 (2/13/95)

  - Communication: 

  - Note: 


********************************************************************************/

#include <cm/cmmd.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <cmsys/cm_fcntl.h>

#include "global.h"

#include "function.h"



/******************************************************************************
 *                                                                            *
 *  Function : read_input()                                                   *
 *                                                                            *
 ******************************************************************************/


void read_input()
{
   Boolean error;
   int i, j;
   int s_total = 1, t_total=1;
   int save;

   CMMD_disable_safety();
   CMMD_fset_io_mode(stdin, CMMD_sync_bc);
   CMMD_fset_io_mode(stdout, CMMD_sync_bc);
   CMMD_fset_io_mode(stderr, CMMD_independent);

   ori_nproc = CMMD_partition_size();

   /*---------------- Rank of the input Array ---------------------------*/
   do {
      printf("\nEnter rank of the input array (Max=%d): ", Rankmax);
      scanf("%d", &rank);
      error = ( rank < 1 || rank > Rankmax );
      if (error)
	 printf("\n??? Improper rank. Try again!\n");
    } while (error);

   /*-------------------- from dim=0 to dim=rank-1 ----------------------*/
   for (i=0; i<rank; ++i) {

       printf("\n***** Dimension = %d *****", i);

       do {
	 printf("\n   Enter domain size (Max=%d): ", Nmax);
	 scanf("%d", &N[i]);
	 error = ( N[i] < 1 || N[i] > Nmax );
	 if (error)
	   printf("\n??? Improper domain size. Try again!\n");
       } while (error);

       /*------------ Source Distribution ------------------*/
       save = s_total;
       do {
	 printf("\n   Enter number of source processors (Max=%d): ", 
                                     Floor(Procmax, s_total));
	 scanf("%d", &s_nproc[i]);
	 s_total *= s_nproc[i];
	 error = ( s_nproc[i] < 1 || 
		   s_total > ori_nproc || s_nproc[i] > N[i] );
	 if (error) {
	   s_total = save;
	   printf("\n??? Improper source processor number. Try again!\n");
	 }
       } while (error);
       
       if ( s_nproc[i] == 1 ) {
	 s_width[i] = N[i];
	 s_L[i]     = N[i];
	 if ( s_L[i] > N_Lmax )
	   CMMD_error("Error: Increas 'N_Lmax' !\n");
       }
       else {
	 do {
	   printf("\n   Enter width of a partition (Max=%d): ", 
                          N[i] / s_nproc[i]);
	   scanf("%d", &s_width[i]);
	   s_L[i] = Ceiling(N[i], s_nproc[i]*s_width[i]) * s_width[i];
	   error = ( s_width[i] < 1 || s_width[i] > s_L[i] || s_L[i] > N_Lmax );
	   if (error)
	     printf("\n??? Improper source partition width. Try again!\n");
	 } while (error);
       }

       /*------------ Target Distribution ------------------*/
       save = t_total;
       do {
	 printf("\n   Enter number of target processors (Max=%d): ", 
                                    Floor(Procmax, t_total));
	 scanf("%d", &t_nproc[i]);
	 t_total *= t_nproc[i];
	 error = ( t_nproc[i] < 1 || 
		   t_total > ori_nproc || t_nproc[i] > N[i] );
	 if (error) {
	   t_total = save;
	   printf("\n??? Improper target processor number. Try again!\n");
	 }
       } while (error);

       if ( t_nproc[i] == 1 ) {
	 t_width[i] = N[i];
	 t_L[i]     = N[i];
	 if ( t_L[i] > N_Lmax )
	   CMMD_error("Error: Increas 'N_Lmax' !\n");
       }
       else {
	 do {
	   printf("\n   Enter width of a partition (Max=%d): ",
		       N[i] / t_nproc[i] );
	   scanf("%d", &t_width[i]);
	   t_L[i] = Ceiling(N[i], t_nproc[i]*t_width[i]) * t_width[i];
	   error = ( t_width[i] < 1 || t_width[i] > t_L[i] || t_L[i] > N_Lmax);
	   if (error)
	     printf("\n??? Improper target partition width. Try again!\n");
	 } while (error);
       }

     }

   /*===== Total number of processors =====*/
   s_tot_nproc = s_total;
   t_tot_nproc = t_total;
 }



/******************************************************************************
 *                                                                            *
 *  Function : initialize()                                                   *
 *                                                                            *
 ******************************************************************************/

void initialize()
{
  int i, j;
   /*===== Clear Counter =====*/
   for (i=0; i<rank; ++i) { /* PITFALL Counter */
     s_PF_count[i] = 0;
     t_PF_count[i] = 0;
     for (j=0; j<ori_nproc; ++j) { /* FALLS Counter */
       s_F_count[i][j] = 0;
       t_F_count[i][j] = 0;
     }
   }

   /*===== Who I am ? =====*/
   my_pnum  = CMMD_self_address(); 
   s_active = my_pnum < s_tot_nproc;
   t_active = my_pnum < t_tot_nproc;

   if (s_active)
     map_to_grid(my_pnum, source, my_s_grid);
   if (t_active)
     map_to_grid(my_pnum, target, my_t_grid);

   /*===== Map-to-Grid =====*/
   for (i=0; i<s_tot_nproc; ++i) 
     map_to_grid(i, source, s_grid[i]);
   for (i=0; i<t_tot_nproc; ++i) 
     map_to_grid(i, target, t_grid[i]);

   /*== ==*/
   for (i=0; i<Tot_Procmax; ++i) 
     sendl[i] = 0;
  
 }

/******************************************************************************
 *                                                                            *
 *  Function : make_PITFALLS()                                                *
 *                                                                            *
 ******************************************************************************/

void make_PITFALLS()
{
  int dim;
  int mod, div, mmoodd, ddiivv;
  int s_cnt, t_cnt;

  /************************************************************
   * NOTE: Regular Distribution has at most 3 PITFALLS        *
   *       Keep track with s_PF_count[dim] & t_PF_count[dim]  *
   ************************************************************/


  /***********************************************************************
   *                                                                     *
   *   N   = div    * twidth + mod                                       *
   *   mod = ddiivv * width  + mmoodd                                    *
   *                                                                     *
   *         div*twidth                                   mod            *
   *  |=======================================|--------------|......|    *
   *         twidth = nproc*width               ddiivv*width  mmoodd     *
   *                                                                     *
   ***********************************************************************/
	

   /*-------------------- from dim=0 to dim=rank-1 ----------------------*/

   for (dim=0; dim<rank; ++dim) {
     
     /*******************************************************************/
     /************************  SOURCE DISTRIBUTION  ********************/
     /*******************************************************************/

     s_cnt = -1;

     s_twidth[dim] = s_width[dim] * s_nproc[dim];
     s_ntile[dim]  = Ceiling(N[dim], s_twidth[dim]);

     div         = N[dim] / s_twidth[dim];
     mod         = N[dim] % s_twidth[dim];
     ddiivv      = mod  / s_width[dim];
     mmoodd      = mod  % s_width[dim];
     
     /*-------------------------------------------------*/
     /*  l  r        str     num  disp   from  to       */
     /* (0, width-1, twidth, div, width, 0,    nproc-1) */
     /*-------------------------------------------------*/

     if ( div > 0 ) {
       ++s_cnt;
       s_PF[dim][s_cnt].l     = 0;
       s_PF[dim][s_cnt].r     = s_width[dim] - 1;
       s_PF[dim][s_cnt].num   = div;

       if ( s_PF[dim][s_cnt].num == 1) 
	 s_PF[dim][s_cnt].str = 1;
       else
	 s_PF[dim][s_cnt].str   = s_twidth[dim];

       if ( s_nproc[dim] > 1 )
	 s_PF[dim][s_cnt].disp  = s_width[dim];
       else
	 s_PF[dim][s_cnt].disp  = 0;
       s_PF[dim][s_cnt].from  = 0;
       s_PF[dim][s_cnt].to    = s_nproc[dim] - 1;
     }

     /*-------------------------------------------------------------------*/
     /*  l           r                   str  num  disp   from  to        */    
     /* (div*twidth, div*twidth+width-1, 1,   1,   width, 0,    ddiivv-1) */
     /*-------------------------------------------------------------------*/

     if ( ddiivv > 0 ) { /* ==> (mod > 0) */
       ++s_cnt;
       s_PF[dim][s_cnt].l    = div * s_twidth[dim];
       s_PF[dim][s_cnt].r    = s_PF[dim][s_cnt].l + s_width[dim] -1;
       s_PF[dim][s_cnt].str  = 1;
       s_PF[dim][s_cnt].num  = 1;
       if ( ddiivv == 1 )
	 s_PF[dim][s_cnt].disp = 0;
       else
	 s_PF[dim][s_cnt].disp = s_width[dim];
       s_PF[dim][s_cnt].from = 0;
       s_PF[dim][s_cnt].to   = ddiivv - 1;
     }

     /*---------------------------------------------------------------------*/
     /*  l                        r                                str  num */ 
     /* (div*twidth+ddiivv*width, div*twidth+ddiivv*width+width-1, 1,   1,  */
     /*                                                                     */
     /*                                              disp   from    to      */ 
     /*                                              width, ddiivv, ddiivv) */
     /*---------------------------------------------------------------------*/

     if ( mmoodd > 0 ) { /* ==> (mod > 0) */
       ++s_cnt;
       s_PF[dim][s_cnt].l    = div * s_twidth[dim] + ddiivv * s_width[dim];
       s_PF[dim][s_cnt].r    = s_PF[dim][s_cnt].l + mmoodd - 1;
       s_PF[dim][s_cnt].str   = s_twidth[dim];
       s_PF[dim][s_cnt].str  = 1;
       s_PF[dim][s_cnt].num  = 1;
       s_PF[dim][s_cnt].disp = 0;
       s_PF[dim][s_cnt].from = ddiivv;
       s_PF[dim][s_cnt].to   = ddiivv;
     }

     s_PF_count[dim] = s_cnt + 1;

     if ( s_PF_count[dim] < 1 ||  s_PF_count[dim] > 3) 
       CMMD_error("Error: Fail to define Source PITFALLS!\n");


     /*******************************************************************/
     /************************  TARGET DISTRIBUTION  ********************/
     /*******************************************************************/

     t_cnt = -1;

     t_twidth[dim] = t_width[dim] * t_nproc[dim];
     t_ntile[dim]  = Ceiling(N[dim], t_twidth[dim]);

     div         = N[dim] / t_twidth[dim];
     mod         = N[dim] % t_twidth[dim];
     ddiivv      = mod  / t_width[dim];
     mmoodd      = mod  % t_width[dim];
     
     /*-------------------------------------------------*/
     /*  l  r        str     num  disp   from  to       */
     /* (0, width-1, twidth, div, width, 0,    nproc-1) */
     /*-------------------------------------------------*/

     if ( div > 0 ) {
       ++t_cnt;
       t_PF[dim][t_cnt].l     = 0;
       t_PF[dim][t_cnt].r     = t_width[dim] - 1;
       t_PF[dim][t_cnt].num   = div;

       if ( t_PF[dim][t_cnt].num == 1 )
	 t_PF[dim][t_cnt].str = 1;
       else
	 t_PF[dim][t_cnt].str = t_twidth[dim];

       if ( t_nproc[dim] > 1 )
	 t_PF[dim][t_cnt].disp  = t_width[dim];
       else
	 t_PF[dim][t_cnt].disp  = 0;
       t_PF[dim][t_cnt].from  = 0;
       t_PF[dim][t_cnt].to    = t_nproc[dim] - 1;
     }

     /*-------------------------------------------------------------------*/
     /*  l           r                   str  num  disp   from  to        */    
     /* (div*twidth, div*twidth+width-1, 1,   1,   width, 0,    ddiivv-1) */
     /*-------------------------------------------------------------------*/

     if ( ddiivv > 0 ) { /* ==> (mod > 0) */
       ++t_cnt;
       t_PF[dim][t_cnt].l    = div * t_twidth[dim];
       t_PF[dim][t_cnt].r    = t_PF[dim][t_cnt].l + t_width[dim] -1;
       t_PF[dim][t_cnt].str  = 1;
       t_PF[dim][t_cnt].num  = 1;
       if ( ddiivv == 1 )
	 t_PF[dim][t_cnt].disp = 0;
       else
	 t_PF[dim][t_cnt].disp = t_width[dim];
       t_PF[dim][t_cnt].from = 0;
       t_PF[dim][t_cnt].to   = ddiivv - 1;
     }

     /*---------------------------------------------------------------------*/
     /*  l                        r                                str  num */ 
     /* (div*twidth+ddiivv*width, div*twidth+ddiivv*width+width-1, 1,   1,  */
     /*                                                                     */
     /*                                              disp   from    to      */ 
     /*                                              width, ddiivv, ddiivv) */
     /*---------------------------------------------------------------------*/

     if ( mmoodd > 0 ) { /* ==> (mod > 0) */
       ++t_cnt;
       t_PF[dim][t_cnt].l    = div * t_twidth[dim] + ddiivv * t_width[dim];
       t_PF[dim][t_cnt].r    = t_PF[dim][t_cnt].l + mmoodd - 1;
       t_PF[dim][t_cnt].str  = 1;
       t_PF[dim][t_cnt].num  = 1;
       t_PF[dim][t_cnt].disp = 0;
       t_PF[dim][t_cnt].from = ddiivv;
       t_PF[dim][t_cnt].to   = ddiivv;
     }

     t_PF_count[dim] = t_cnt + 1;

     if ( t_PF_count[dim] < 1 || t_PF_count[dim] > 3)
       CMMD_error("Error: Fail to define Target PITFALLS!\n");

   } /* FOR */
   

  CMMD_node_timer_stop(0); 

   printf("\n**************************************************************");
   printf("\n                  INPUT DATA (Rank = %d)", rank);
   printf("\n**************************************************************\n");
   printf("\n Total number of processors: Source = %d", s_tot_nproc);
   printf("\n                             Target = %d\n", t_tot_nproc);
   printf("\n**************************************************************\n");
   for (dim=0; dim<rank; ++dim) {
     printf("\n <<< Dimension = %d >>>", dim);
     printf("\n Domain size            = %d\n", N[dim]);

     printf("\n Source Processors      = %d", s_nproc[dim]);
     printf("\n        Partition Width = %d", s_width[dim]);
     printf("\n        Local domain    = %d", s_L[dim]);
     printf("\n        PITFALLS Count  = %d", s_PF_count[dim]);
     for (s_cnt=0; s_cnt<s_PF_count[dim]; ++s_cnt)
       printf("\n                 #%2d    = (%3d, %3d, %3d, %3d, %3d, %3d, %3d)",
	      s_cnt, s_PF[dim][s_cnt].l, s_PF[dim][s_cnt].r, 
	         s_PF[dim][s_cnt].str, s_PF[dim][s_cnt].num,
	         s_PF[dim][s_cnt].disp, s_PF[dim][s_cnt].from,
                 s_PF[dim][s_cnt].to);

     printf("\n\n Target Processors      = %d", t_nproc[dim]);
     printf("\n        Partition Width = %d", t_width[dim]);
     printf("\n        Local domain    = %d", t_L[dim]);
     printf("\n        PITFALLS Count  = %d", t_PF_count[dim]);
     for (t_cnt=0; t_cnt<t_PF_count[dim]; ++t_cnt)
       printf("\n                 #%2d    = (%3d, %3d, %3d, %3d, %3d, %3d, %3d)",
	      t_cnt, t_PF[dim][t_cnt].l, t_PF[dim][t_cnt].r, 
	         t_PF[dim][t_cnt].str, t_PF[dim][t_cnt].num,
	         t_PF[dim][t_cnt].disp, t_PF[dim][t_cnt].from,
                 t_PF[dim][t_cnt].to);
     printf("\n**************************************************************\n");
   }

  CMMD_node_timer_start(0); 
}


/******************************************************************************
 *                                                                            *
 *  Function : void LS_intersection(L1, L2, LI)                               *
 *                                                                            *
 ******************************************************************************/

void LS_intersection(L1, L2, LI)
     LS L1, L2;
     LS *LI;
{
  int mx = max(L1.l, L2.l), 
      mn = min(L1.r, L2.r);

  if ( mx <= mn ) {
    LI->l = mx;
    LI->r = mn;
  }
  else {
    LI->l = empty;
    LI->r = empty;
  }
}



/******************************************************************************
 *                                                                            *
 *  Function : Boolean Empty_FALLS_intersection(F1, F2);                      *
 *                                                                            *
 ******************************************************************************/

/*---------------------------
     double ceil(double x);
     double floor(double x);
  ---------------------------*/

Boolean Empty_FALLS_intersection(F1, F2)
     FALLS F1, F2;
{
  int i1, i2;
  int I1, I2;
  LS  L1, L2;

  int l1 = F1.l, r1 = F1.r, num1 = F1.num, str1 = F1.str;
  int l2 = F2.l, r2 = F2.r, num2 = F2.num, str2 = F2.str;

  int period = lcm(str1, str2);

  int count = 0;

  int m1, m2;

  if ( num1 == 1 || num2 ==1 ) {
    m1 = num1;
    m2 = num2;
  }
  else {
    m1 = period/str1;
    m2 = period/str2;
  }


  I1 = max( 0, (int)(ceil( (double)(l2-r1) / (double)(str1) )) );

  for (i1 = I1; i1 <= min(I1 + m1 - 1, num1 - 1); ++i1)
    for (i2 = max(0, (int)(ceil( (double)(i1*str1+l1-r2) / (double)(str2) )) );

	 i2 <= min( 
	       min( (int)(floor( (double)(i1*str1+r1-l2) / (double)(str2) )), 
                    m2-1),
              num2-1  );
	++i2) 
      return false;

  return true;
}



/******************************************************************************
 *                                                                            *
 *  Function : void FALLS_intersection(F1, F2, FI );                          *
 *                                                                            *
 ******************************************************************************/

void FALLS_intersection(F1, F2, FI)
     FALLS F1, F2;
     Queue *FI;
{
  int i1, i2;
  int I1, I2;
  LS  L1, L2, LI;
  
  FALLS new;

  int l1 = F1.l, r1 = F1.r, num1 = F1.num, str1 = F1.str;
  int l2 = F2.l, r2 = F2.r, num2 = F2.num, str2 = F2.str;

  int period = lcm(str1, str2);

  int m1, m2;

  if ( num1 == 1 || num2 ==1 ) {
    m1 = num1;
    m2 = num2;
  }
  else {
    m1 = period/str1;
    m2 = period/str2;
  }

  /*clearQ(FI);*/

  I1 = max( 0, (int)(ceil( (double)(l2-r1) / (double)(str1) )) );

  for (i1 = I1; i1 <= min(I1 + m1 - 1, num1 - 1); ++i1) {
    L1.l = l1+i1*str1;
    L1.r = r1+i1*str1;

    for (i2 = max(0, (int)(ceil( (double)(i1*str1+l1-r2) / (double)(str2) )) );

	 i2 <= min( 
	       min( (int)(floor( (double)(i1*str1+r1-l2) / (double)(str2) )), 
                    m2-1),
              num2-1  );
	++i2) {
          L2.l = l2+i2*str2;
	  L2.r = r2+i2*str2;
	  LS_intersection(L1, L2, &LI);
	  new.l = LI.l;
	  new.r = LI.r;
	  new.num = min( (num1-i1-1)/m1, (num2-i2-1)/m2 ) + 1;
	  if (new.num == 1)
	    new.str = 1;     /* Block distribution */
	  else
	    new.str = period;
	  enQ(FI, new);
	}
  }

}


/******************************************************************************
 *                                                                            *
 *  Function : make_FALLS();                                                  *
 *                                                                            *
 ******************************************************************************/

void make_FALLS()
{
  Pindex s_cnt, t_cnt, s_pn, t_pn, s_c, t_c;
  int dim, i;    

/*FOR DEBUG*/ int k;

  /************************************************************/
  /* NOTE: Each processor has at most 2 FALLS                 */
  /*       Keep track with s_F_count[][] & tPF_count[][]      */
  /************************************************************/

  for (dim=0; dim<rank; ++dim) {

    /*------------------------- Source ---------------------------*/
    for (s_cnt=0; s_cnt<s_PF_count[dim]; ++s_cnt) {

      for (s_pn =  s_PF[dim][s_cnt].from; 
	   s_pn <= s_PF[dim][s_cnt].to;   ++s_pn) {

	/* define FALLS for source processor "s_pn" */
	
	s_c = s_F_count[dim][s_pn];
	++ s_F_count[dim][s_pn];
	if (s_F_count[dim][s_pn] > 2) 
	  CMMD_error("Error: Fail to define Source FALLS!\n");

	    
	s_F[dim][s_pn][s_c].l   = s_PF[dim][s_cnt].l + 
	                          s_pn * s_PF[dim][s_cnt].disp;
	s_F[dim][s_pn][s_c].r   = s_PF[dim][s_cnt].r + 
	                          s_pn * s_PF[dim][s_cnt].disp;
	s_F[dim][s_pn][s_c].num = s_PF[dim][s_cnt].num;
	s_F[dim][s_pn][s_c].str = s_PF[dim][s_cnt].str;
      } /* s_pn */
    } /* s_cnt */

/*FOR DEBUG*/ k=4;

    /*------------------------- Target ---------------------------*/
    for (t_cnt=0; t_cnt<t_PF_count[dim]; ++t_cnt) {

      for (t_pn =  t_PF[dim][t_cnt].from; 
	   t_pn <= t_PF[dim][t_cnt].to;   ++t_pn) {

	/* define FALLS for source processor "t_pn" */
	
	t_c = t_F_count[dim][t_pn];
	++ t_F_count[dim][t_pn];
	if (t_F_count[dim][t_pn] > 2)
	  CMMD_error("Error: Fail to define Target FALLS!\n");
	    
	t_F[dim][t_pn][t_c].l   = t_PF[dim][t_cnt].l + 
	                          t_pn * t_PF[dim][t_cnt].disp;
	t_F[dim][t_pn][t_c].r   = t_PF[dim][t_cnt].r + 
	                          t_pn * t_PF[dim][t_cnt].disp;
	t_F[dim][t_pn][t_c].num = t_PF[dim][t_cnt].num;
	t_F[dim][t_pn][t_c].str = t_PF[dim][t_cnt].str;

      } /* t_pn */
    } /* t_cnt */

/*FOR DEBUG*/ k=4;

    /*----------------------- Define my_s_F ---------------------*/
    if ( s_active ) {
      my_s_F_count[dim] = s_F_count[dim][my_s_grid[dim]];
      for (i=0; i<s_F_count[dim][my_s_grid[dim]]; ++i) {
	my_s_F[dim][i].l   = s_F[dim][my_s_grid[dim]][i].l;
	my_s_F[dim][i].r   = s_F[dim][my_s_grid[dim]][i].r;
	my_s_F[dim][i].str = s_F[dim][my_s_grid[dim]][i].str;
	my_s_F[dim][i].num = s_F[dim][my_s_grid[dim]][i].num;
      }
    }

    else
      my_s_F_count[dim] = 0;

    /*---------------------- Define my_t_F ----------------------*/
    if ( t_active ) {
      my_t_F_count[dim] = t_F_count[dim][my_t_grid[dim]];
      for (i=0; i<t_F_count[dim][my_t_grid[dim]]; ++i) {
	my_t_F[dim][i].l   = t_F[dim][my_t_grid[dim]][i].l;
	my_t_F[dim][i].r   = t_F[dim][my_t_grid[dim]][i].r;
	my_t_F[dim][i].str = t_F[dim][my_t_grid[dim]][i].str;
	my_t_F[dim][i].num = t_F[dim][my_t_grid[dim]][i].num;
      }
    }

    else
      my_t_F_count[dim] = 0;

  } /* dim */

}



/******************************************************************************
 *                                                                            *
 *  Function : PITFALLS_intersection()                                        *
 *                                                                            *
 ******************************************************************************/

/* Distributed Version using make_FALLS() */

void PITFALLS_intersection()
{
  Pindex s_pn, t_pn, s_c, t_c;
  int dim;    

/*FOR DEBUG*/ int k; 

  /************************************************************/
  /* NOTE: Each processor has at most 2 FALLS                 */
  /*       Keep track with s_F_count[][] & tPF_count[][]      */
  /************************************************************/

  for (dim=0; dim<rank; ++dim) {

    /* For sending */
    for (t_pn=0; t_pn<t_nproc[dim]; ++t_pn) {

      clearQ(&send_FI[dim][t_pn]);  /* Clear queue */

      for (s_c = 0; s_c < my_s_F_count[dim]; ++s_c)

	for (t_c = 0; t_c < t_F_count[dim][t_pn]; ++t_c)

	  if ( ! Empty_FALLS_intersection(
				 my_s_F[dim][s_c], t_F[dim][t_pn][t_c] ) )
	    FALLS_intersection( my_s_F[dim][s_c], t_F[dim][t_pn][t_c],
				  &send_FI[dim][t_pn] );

      /*FOR DEBUG*/ k=4;
    } /* t_pn */

    /* For receiving */
    for (s_pn=0; s_pn<s_nproc[dim]; ++s_pn) {

      clearQ(&recv_FI[dim][s_pn]);  /* Clear queue */

      for (s_c = 0; s_c < s_F_count[dim][s_pn]; ++s_c)

	for (t_c = 0; t_c < my_t_F_count[dim]; ++t_c)

	  if ( ! Empty_FALLS_intersection(
				 s_F[dim][s_pn][s_c], my_t_F[dim][t_c] ) )
	    FALLS_intersection( s_F[dim][s_pn][s_c], my_t_F[dim][t_c],
				  &recv_FI[dim][s_pn] );

      /*FOR DEBUG*/ k=4;
    } /* s_pn */


  } /* dim */

}



/******************************************************************************
 *                                                                            *
 *  Function : int get_index(FI, Global-index)                                *
 *                                                                            *
 ******************************************************************************/

int get_index(q, lindex, st, dim)
     Queue *q;
     Lindex lindex[Indexmax];
     int st;
     int dim;
{
  FALLS fi;
  Gindex gindex;
  int size = 0, save, num=0, i, j, k, dif;

  save = q->size;

  if ( save > 0 ) {
    k = 0;
    while (! emptyQ(q)) {
      fi = deQ(q);
      dif = fi.r - fi.l + 1;
      num += dif * fi.num;
      for (i=0; i<fi.num; ++i) 
	for (j=fi.l; j<=fi.r; ++j) {
	  gindex = j + i*fi.str;
	  lindex[k] = map_to_local(st, gindex, dim);
	  ++k;
	}
    }
/*********************
    resetQ(q, save);
**********************/
    if (num != k)
      CMMD_error("Error: Fail to generating index set!\n");

    return num;
  }

  else
    return 0;

}
  


/******************************************************************************
 *                                                                            *
 *  Function : void communication_matrix()                                    *
 *                                                                            *
 ******************************************************************************/

void communication_matrix() 
{
  int sp, sP, tp, tP, d, num, g, i, offset;

  /*===== For each dimension, get local index and counter for sending =====*/
  for (d=0; d<rank; ++d) 
    for (tp=0; tp<t_nproc[d]; ++tp) 
      send_index_count[d][tp] = get_index( &send_FI[d][tp], 
                                           send_Lindex[d][tp],
					   source,
                                           d);

  /*=== sendl vector ===*/
  offset = 0;
  for (tP=0; tP<t_tot_nproc; ++tP) {
    sendl[tP] = 1;
    for (d=0; d<rank; ++d) {
      num = send_index_count[d][ t_grid[tP][d] ];
      if ( num == 0 ) {/* empty */
	sendl[tP] = 0;
	break;
      }
      sendl[tP] *= num;
    } /* d */
    send_offset[tP] = offset;
    offset += sendl[tP];
    if ( sendl[tP] > Bufmax ) 
      CMMD_error("Error: Overflow Send Buffer !\n");
  } /* tP */


  /*===== For each dimension, get local index and counter for receiving =====*/
  for (d=0; d<rank; ++d) 
    for (sp=0; sp<s_nproc[d]; ++sp) 
      recv_index_count[d][sp] = get_index( &recv_FI[d][sp], 
					   recv_Lindex[d][sp],
					   target,
					   d);

  /*=== recvl vector ===*/
/********
  ----------------------------------------------------------------------
    recvl[] is calculated in one-stage communication 
    (= extern int recd_msg_len[])
    recv_offset[] is calculated in one-stage communication
  ----------------------------------------------------------------------
  offset=0;
  for (sP=0; sP<s_tot_nproc; ++sP) {
    recvl[sP] = 1;
    for (d=0; d<rank; ++d) {
      num = recv_index_count[d][ s_grid[sP][d] ];
      if ( num == 0 ) {
	recvl[sP] = 0;
	break;
      }
      recvl[sP] *= num;
    } 

    recv_offset[sP] = offset;
    offset += recvl[sP];

    if ( recvl[sP] > Bufmax ) 
      CMMD_error("Error: Overflow Receive Buffer !\n");
  } 

  ----------------------------------------------------------------------
*******/

}



/******************************************************************************
 *                                                                            *
 *  Function : void compose-message()                                         *
 *                                                                            *
 ******************************************************************************/

void compose_message()
{
  int tp;
  void comp_msg1(), comp_msg2(), comp_msg3(), comp_msg4(),comp_msg5(); 

  switch (rank) {
     case 1: comp_msg1(); 
             break;
     case 2: comp_msg2();
             break;
     case 3: comp_msg3();
             break;
     case 4: comp_msg4();
             break;
     case 5: comp_msg5();
             break;
     default:CMMD_error("Error: Sorry, we can not support the array rank larger than %d!\n", Rankmax);
     }

}



/******************************************************************************
 *                                                                            *
 *  Function : void comp_msg*()                                               *
 *                                                                            *
 ******************************************************************************/

void comp_msg1()
{
  int tp, i, b;

  for (tp=0; tp<t_tot_nproc; ++tp) 
    if ( sendl[tp] ) {
      for (i=0; i<send_index_count[0][tp]; ++i) 
	sendB[ send_offset[tp] + i] = A1[ send_Lindex[0][tp][i] ];
    }
  
}


void comp_msg2()
{
  int tp, i0, i1, b;

  for (tp=0; tp<t_tot_nproc; ++tp) 
    if ( sendl[tp] ) {
      b = 0;
      for (i1=0; i1<send_index_count[1][ t_grid[tp][1] ]; ++i1)
	for (i0=0; i0<send_index_count[0][ t_grid[tp][0] ]; ++i0) {
	sendB[ send_offset[tp] + b ] = 
               A2[ send_Lindex[1][ t_grid[tp][1] ][i1] ]
	         [ send_Lindex[0][ t_grid[tp][0] ][i0] ];
	++b;
      }
    }
}


void comp_msg3()
{
}

void comp_msg4()
{
}

void comp_msg5()
{
}



/******************************************************************************
 *                                                                            *
 *  Function : distributed_scheduling()                                       *
 *                                                                            *
 ******************************************************************************/


void distributed_scheduling()
{

  Initialization(); 

  one_stage(sendB, recvB, sendl, send_offset);

}

/******************************************************************************
 *                                                                            *
 *  Function : void decompose-message()                                       *
 *                                                                            *
 ******************************************************************************/

void decompose_message()
{
  int sp;
  void decomp_msg1(), decomp_msg2(), decomp_msg3(), decomp_msg4(),decomp_msg5(); 

  switch (rank) {
     case 1: decomp_msg1(); 
             break;
     case 2: decomp_msg2();
             break;
     case 3: decomp_msg3();
             break;
     case 4: decomp_msg4();
             break;
     case 5: decomp_msg5();
             break;
     default:CMMD_error("Error: Sorry, we can not support the array rank larger than %d!\n", Rankmax);
     }

}


/******************************************************************************
 *                                                                            *
 *  Function : void decomp_msg*()                                             *
 *                                                                            *
 ******************************************************************************/


void decomp_msg1()
{
  int sp, m, i;

  for (m=0; m<recd_msg_no; ++m) {
    sp = recd_msg_source[m];
    for (i=0; i<recv_index_count[0][sp]; ++i) 
      A1[ recv_Lindex[0][sp][i] ] = recvB[ recv_offset[sp] + i ];
  }

}	




void decomp_msg2()
{
  int sp, m, i0, i1, b;

  for (m=0; m<recd_msg_no; ++m) {
    sp = recd_msg_source[m];
    b = 0;
    for (i1=0; i1<recv_index_count[1][ s_grid[sp][1] ]; ++i1)
      for (i0=0; i0<recv_index_count[0][ s_grid[sp][0] ]; ++i0) {
	A2[ recv_Lindex[1][ s_grid[sp][1] ][i1] ]
	  [ recv_Lindex[0][ s_grid[sp][0] ][i0] ]
	    = recvB[ recv_offset[sp] + b ];
	++b;
      }
  }

}

void decomp_msg3()
{
}

void decomp_msg4()
{
}

void decomp_msg5()
{
}



/******************************************************************************
 *                                                                            *
 *  Function : make_array*()                                                  *
 *                                                                            *
 ******************************************************************************/

void make_array1()
{  
  int i, /* local */
      r; /* global */

  for (i=0; i<s_L[0]; ++i) {
    r = map_to_global(source, my_pnum, i, 0);
    A1[i] = r;
  }

}


void make_array2()
{
  int r, c,     /* global index */
      i, j;     /* local index  */

  for (i=0; i<s_L[1]; ++i) {
    r = map_to_global(source, my_s_grid[1], i, 1);
    for (j=0; j<s_L[0]; ++j) {
      c = map_to_global(source, my_s_grid[0], j, 0);
      A2[i][j] = r*N[0] + c;
    }
  }

}
  
/******************************************************************************
 *                                                                            *
 *  Function : print_array*()                                                 *
 *                                                                            *
 ******************************************************************************/

void print_array1(st)
     int st;
{
  int r, /*global*/
      i; /*local*/
  int pn;
  int ele;

  CMMD_sync_with_nodes();
  CMMD_fset_io_mode(stdout,CMMD_sync_bc);
  if (st == source)
    printf("\n\n ******************** SOURCE ARRAY ********************\n\n"); 
  else
    printf("\n\n ******************** TARGET ARRAY ********************\n\n"); 
  CMMD_fset_io_mode(stdout, CMMD_independent);

  CMMD_sync_with_nodes(); 
  CMMD_fset_io_mode(stdout, CMMD_independent); 

  for (r=0; r<N[0]; ++r) {
    CMMD_sync_with_nodes(); 
    
    i = map_to_local(st, r, 0);
    pn = map_to_proc(st, r, 0);

    ele = A1[i];

    if ( my_pnum == 0 ) {

      if ( pn == 0 )
	printf("%7d", ele);
      else {
	CMMD_receive_block(pn, CMMD_ANY_TAG, &ele, 4);
	printf("%7d", ele);
      };
    }
    else if ( my_pnum == pn ) 
      CMMD_send_block(0, CMMD_DEFAULT_TAG, &ele, 4);

  }

  if (my_pnum == 0 )
    printf("\n");
       
  CMMD_sync_with_nodes();
  CMMD_fset_io_mode(stdout,CMMD_sync_bc);
  printf("\n");
  CMMD_fset_io_mode(stdout, CMMD_independent);

}

void print_array2(st)
     int st;
{
   int r, c,     /* global index */
       i, j,     /* local index  */
       pn, grid[Rankmax];
   int ele;
 

   CMMD_sync_with_nodes();
   CMMD_fset_io_mode(stdout,CMMD_sync_bc);
   if (st == source)
     printf("\n\n ******************** SOURCE ARRAY ********************\n\n"); 
   else
     printf("\n\n ******************** TARGET ARRAY ********************\n\n"); 
   CMMD_fset_io_mode(stdout, CMMD_independent);

   CMMD_sync_with_nodes(); 
   CMMD_fset_io_mode(stdout, CMMD_independent); 

   for (r=0; r<N[1]; ++r) {
      CMMD_sync_with_nodes(); 
      
      i = map_to_local(st, r, 1);
      grid[1] = map_to_proc(st, r, 1);

      for (c=0; c<N[0]; ++c) {  
         CMMD_sync_with_nodes();

         j  = map_to_local(st, c, 0);
         grid[0] = map_to_proc(st, c, 0);

         map_to_pnum(grid, st, &pn);

         ele = A2[i][j];

         CMMD_sync_with_nodes();
         if ( my_pnum == 0 ) {

            if ( pn == 0 )
              printf("%7d", ele);
            else {
              CMMD_receive_block(pn, CMMD_ANY_TAG, &ele, 4);
              printf("%7d", ele);
           };
         }
         else if ( my_pnum == pn ) 
             CMMD_send_block(0, CMMD_DEFAULT_TAG, &ele, 4);
       
      }; /* for(c) */

         if (my_pnum == 0 ) {
              printf("\n");
           };
      
   }; /* for(r) */

   CMMD_sync_with_nodes();
   CMMD_fset_io_mode(stdout,CMMD_sync_bc);
   printf("\n");
   CMMD_fset_io_mode(stdout, CMMD_independent);
 };


/******************************************************************************
 *                                                                            *
 *  Function : main()                                                         *
 *                                                                            *
 ******************************************************************************/

main()
{
  int p,q,l,i,d;

  double total, run, comm, s_comp, r_comp, max, min, ave;

  read_input();

  CMMD_node_timer_clear(0);  /* s_comp */
  CMMD_node_timer_clear(1);  /* r_comp */
  CMMD_node_timer_clear(2);  /* comm */

  CMMD_sync_with_nodes();

  CMMD_node_timer_start(0);   /* s_comp */

  initialize();

/*****
  CMMD_node_timer_stop(0); 
  if ( rank == 1 )
    make_array1();
  else if ( rank == 2 )
    make_array2();
  CMMD_node_timer_start(0); 
*****/

  make_PITFALLS();

  make_FALLS();

  PITFALLS_intersection();

  communication_matrix();

  CMMD_node_timer_stop(0);    /* s_comp */

  CMMD_node_timer_start(1);   /* r_comp */

  compose_message();

  CMMD_node_timer_stop(1);    /* r_comp */

  CMMD_node_timer_start(2);   /* comm */

  /*unschedule_cmmd_async();*/
  /*linear_permutation_cmmd();*/
  distributed_scheduling();

  CMMD_node_timer_stop(2);    /* comm */

  CMMD_node_timer_start(1);   /* r_comp */

  decompose_message();

  CMMD_node_timer_stop(1);    /* r_comp */

  s_comp = CMMD_node_timer_busy(0) * 1000.;  /* milisec */
  r_comp = CMMD_node_timer_busy(1) * 1000.;  /* milisec */
  comm   = CMMD_node_timer_busy(2) * 1000.;  /* milisec */

  total = s_comp + r_comp + comm;
  run   = r_comp + comm;
  max  = CMMD_reduce_double(total, CMMD_combiner_dmax);
  min  = CMMD_reduce_double(total, CMMD_combiner_dmin);
  ave  = CMMD_reduce_double(total, CMMD_combiner_dadd) / 32;

  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n\n***************************************");
  printf("\n          Timining Results ");
  printf("\n***************************************\n\n");
  printf("  Max Total     = %6.2f msec\n", max);
  printf("  Min Total     = %6.2f msec\n", min);
  printf("  Ave Total     = %6.2f msec\n\n", ave);
  printf("         <<< Processor 0 >>>\n");
  printf("  Total         = %6.2f msec\n", total);
  printf("  Detection     = %6.2f msec (%4.1f%%)\n", s_comp, s_comp/total*100.);
  printf("  Comp./Decomp. = %6.2f msec (%4.1f%%)\n", r_comp, r_comp/total*100.);
  printf("  Communication = %6.2f msec (%4.1f%%)\n", comm,   comm/total*100.);
  printf("  Run time      = %6.2f msec (%4.1f%%)\n", run,    run/total*100.);
  printf("\n***************************************\n\n");

/*printf(" %6.2f & %6.2f (%.1f%%) & %6.2f (%.1f%%) & %6.2f (%.1f%%) \\\\ \n",
   total, s_comp, s_comp/total*100., 
   r_comp, r_comp/total*100., comm, comm/total*100.);
printf("\n %6.2f & %6.2f \\\\ \n\n", total, run);*/

printf(" %6.2f & %6.2f (%.1f$\\%%$) & %6.2f (%.1f$\\%%$) & %6.2f (%.1f$\\%%$) \\\\ \n",
   total, s_comp, s_comp/total*100., 
   r_comp, r_comp/total*100., comm, comm/total*100.);
printf("\n %6.2f & %6.2f \\\\ \n\n", total, run);

/*****
  if ( rank == 1 ) 
    print_array1(target);
  else if ( rank == 2 ) 
    print_array2(target);
*****/
}
